use std::{
    fmt::{Display, Formatter},
    iter::Peekable,
    str::CharIndices,
};

// Wrapper to avoid unnecessary branching when input doesn't have ANSI escape sequences.
pub struct AnsiStyle {
    attributes: Option<Attributes>,
}

impl AnsiStyle {
    pub fn new() -> Self {
        AnsiStyle { attributes: None }
    }

    pub fn update(&mut self, sequence: &str) -> bool {
        match &mut self.attributes {
            Some(a) => a.update(sequence),
            None => {
                self.attributes = Some(Attributes::new());
                self.attributes.as_mut().unwrap().update(sequence)
            }
        }
    }
}

impl Display for AnsiStyle {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        match self.attributes {
            Some(ref a) => a.fmt(f),
            None => Ok(()),
        }
    }
}

struct Attributes {
    foreground: String,
    background: String,
    underlined: String,

    /// The character set to use.
    /// REGEX: `\^[()][AB0-3]`
    charset: String,

    /// A buffer for unknown sequences.
    unknown_buffer: String,

    /// ON:  ^[1m
    /// OFF: ^[22m
    bold: String,

    /// ON:  ^[2m
    /// OFF: ^[22m
    dim: String,

    /// ON:  ^[4m
    /// OFF: ^[24m
    underline: String,

    /// ON:  ^[3m
    /// OFF: ^[23m
    italic: String,

    /// ON:  ^[9m
    /// OFF: ^[29m
    strike: String,
}

impl Attributes {
    pub fn new() -> Self {
        Attributes {
            foreground: "".to_owned(),
            background: "".to_owned(),
            underlined: "".to_owned(),
            charset: "".to_owned(),
            unknown_buffer: "".to_owned(),
            bold: "".to_owned(),
            dim: "".to_owned(),
            underline: "".to_owned(),
            italic: "".to_owned(),
            strike: "".to_owned(),
        }
    }

    /// Update the attributes with an escape sequence.
    /// Returns `false` if the sequence is unsupported.
    pub fn update(&mut self, sequence: &str) -> bool {
        let mut chars = sequence.char_indices().skip(1);

        if let Some((_, t)) = chars.next() {
            match t {
                '(' => self.update_with_charset('(', chars.map(|(_, c)| c)),
                ')' => self.update_with_charset(')', chars.map(|(_, c)| c)),
                '[' => {
                    if let Some((i, last)) = chars.last() {
                        // SAFETY: Always starts with ^[ and ends with m.
                        self.update_with_csi(last, &sequence[2..i])
                    } else {
                        false
                    }
                }
                _ => self.update_with_unsupported(sequence),
            }
        } else {
            false
        }
    }

    fn sgr_reset(&mut self) {
        self.foreground.clear();
        self.background.clear();
        self.underlined.clear();
        self.bold.clear();
        self.dim.clear();
        self.underline.clear();
        self.italic.clear();
        self.strike.clear();
    }

    fn update_with_sgr(&mut self, parameters: &str) -> bool {
        let mut iter = parameters
            .split(';')
            .map(|p| if p.is_empty() { "0" } else { p })
            .map(|p| p.parse::<u16>())
            .map(|p| p.unwrap_or(0)); // Treat errors as 0.

        while let Some(p) = iter.next() {
            match p {
                0 => self.sgr_reset(),
                1 => self.bold = format!("\x1B[{}m", parameters),
                2 => self.dim = format!("\x1B[{}m", parameters),
                3 => self.italic = format!("\x1B[{}m", parameters),
                4 => self.underline = format!("\x1B[{}m", parameters),
                23 => self.italic.clear(),
                24 => self.underline.clear(),
                22 => {
                    self.bold.clear();
                    self.dim.clear();
                }
                30..=39 => self.foreground = Self::parse_color(p, &mut iter),
                40..=49 => self.background = Self::parse_color(p, &mut iter),
                58..=59 => self.underlined = Self::parse_color(p, &mut iter),
                90..=97 => self.foreground = Self::parse_color(p, &mut iter),
                100..=107 => self.foreground = Self::parse_color(p, &mut iter),
                _ => {
                    // Unsupported SGR sequence.
                    // Be compatible and pretend one just wasn't was provided.
                }
            }
        }

        true
    }

    fn update_with_csi(&mut self, finalizer: char, sequence: &str) -> bool {
        if finalizer == 'm' {
            self.update_with_sgr(sequence)
        } else {
            false
        }
    }

    fn update_with_unsupported(&mut self, sequence: &str) -> bool {
        self.unknown_buffer.push_str(sequence);
        false
    }

    fn update_with_charset(&mut self, kind: char, set: impl Iterator<Item = char>) -> bool {
        self.charset = format!("\x1B{}{}", kind, set.take(1).collect::<String>());
        true
    }

    fn parse_color(color: u16, parameters: &mut dyn Iterator<Item = u16>) -> String {
        match color % 10 {
            8 => match parameters.next() {
                Some(5) /* 256-color */ => format!("\x1B[{};5;{}m", color, join(";", 1, parameters)),
                Some(2) /* 24-bit color */ => format!("\x1B[{};2;{}m", color, join(";", 3, parameters)),
                Some(c) => format!("\x1B[{};{}m", color, c),
                _ => "".to_owned(),
            },
            9 => "".to_owned(),
            _ => format!("\x1B[{}m", color),
        }
    }
}

impl Display for Attributes {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}{}{}{}{}{}{}{}{}",
            self.foreground,
            self.background,
            self.underlined,
            self.charset,
            self.bold,
            self.dim,
            self.underline,
            self.italic,
            self.strike,
        )
    }
}

fn join(
    delimiter: &str,
    limit: usize,
    iterator: &mut dyn Iterator<Item = impl ToString>,
) -> String {
    iterator
        .take(limit)
        .map(|i| i.to_string())
        .collect::<Vec<String>>()
        .join(delimiter)
}

/// A range of indices for a raw ANSI escape sequence.
#[derive(Debug, PartialEq)]
enum EscapeSequenceOffsets {
    Text {
        start: usize,
        end: usize,
    },
    Unknown {
        start: usize,
        end: usize,
    },
    NF {
        // https://en.wikipedia.org/wiki/ANSI_escape_code#nF_Escape_sequences
        start_sequence: usize,
        start: usize,
        end: usize,
    },
    OSC {
        // https://en.wikipedia.org/wiki/ANSI_escape_code#OSC_(Operating_System_Command)_sequences
        start_sequence: usize,
        start_command: usize,
        start_terminator: usize,
        end: usize,
    },
    CSI {
        // https://en.wikipedia.org/wiki/ANSI_escape_code#CSI_(Control_Sequence_Introducer)_sequences
        start_sequence: usize,
        start_parameters: usize,
        start_intermediates: usize,
        start_final_byte: usize,
        end: usize,
    },
}

/// An iterator over the offests of ANSI/VT escape sequences within a string.
///
/// ## Example
///
/// ```ignore
/// let iter = EscapeSequenceOffsetsIterator::new("\x1B[33mThis is yellow text.\x1B[m");
/// ```
struct EscapeSequenceOffsetsIterator<'a> {
    text: &'a str,
    chars: Peekable<CharIndices<'a>>,
}

impl<'a> EscapeSequenceOffsetsIterator<'a> {
    pub fn new(text: &'a str) -> EscapeSequenceOffsetsIterator<'a> {
        return EscapeSequenceOffsetsIterator {
            text,
            chars: text.char_indices().peekable(),
        };
    }

    /// Takes values from the iterator while the predicate returns true.
    /// If the predicate returns false, that value is left.
    fn chars_take_while(&mut self, pred: impl Fn(char) -> bool) -> Option<(usize, usize)> {
        if self.chars.peek().is_none() {
            return None;
        }

        let start = self.chars.peek().unwrap().0;
        let mut end: usize = start;
        while let Some((i, c)) = self.chars.peek() {
            if !pred(*c) {
                break;
            }

            end = *i + c.len_utf8();
            self.chars.next();
        }

        Some((start, end))
    }

    fn next_text(&mut self) -> Option<EscapeSequenceOffsets> {
        match self.chars_take_while(|c| c != '\x1B') {
            None => None,
            Some((start, end)) => Some(EscapeSequenceOffsets::Text { start, end }),
        }
    }

    fn next_sequence(&mut self) -> Option<EscapeSequenceOffsets> {
        let (start_sequence, c) = self.chars.next().expect("to not be finished");
        match self.chars.peek() {
            None => Some(EscapeSequenceOffsets::Unknown {
                start: start_sequence,
                end: start_sequence + c.len_utf8(),
            }),

            Some((_, ']')) => self.next_osc(start_sequence),
            Some((_, '[')) => self.next_csi(start_sequence),
            Some((i, c)) => match c {
                '\x20'..='\x2F' => self.next_nf(start_sequence),
                c => Some(EscapeSequenceOffsets::Unknown {
                    start: start_sequence,
                    end: i + c.len_utf8(),
                }),
            },
        }
    }

    fn next_osc(&mut self, start_sequence: usize) -> Option<EscapeSequenceOffsets> {
        let (osc_open_index, osc_open_char) = self.chars.next().expect("to not be finished");
        debug_assert_eq!(osc_open_char, ']');

        let mut start_terminator: usize;
        let mut end_sequence: usize;

        loop {
            match self.chars_take_while(|c| !matches!(c, '\x07' | '\x1B')) {
                None => {
                    start_terminator = self.text.len();
                    end_sequence = start_terminator;
                    break;
                }

                Some((_, end)) => {
                    start_terminator = end;
                    end_sequence = end;
                }
            }

            match self.chars.next() {
                Some((ti, '\x07')) => {
                    end_sequence = ti + '\x07'.len_utf8();
                    break;
                }

                Some((ti, '\x1B')) => {
                    match self.chars.next() {
                        Some((i, '\\')) => {
                            end_sequence = i + '\\'.len_utf8();
                            break;
                        }

                        None => {
                            end_sequence = ti + '\x1B'.len_utf8();
                            break;
                        }

                        _ => {
                            // Repeat, since `\\`(anything) isn't a valid ST.
                        }
                    }
                }

                None => {
                    // Prematurely ends.
                    break;
                }

                Some((_, tc)) => {
                    panic!("this should not be reached: char {:?}", tc)
                }
            }
        }

        Some(EscapeSequenceOffsets::OSC {
            start_sequence,
            start_command: osc_open_index + osc_open_char.len_utf8(),
            start_terminator: start_terminator,
            end: end_sequence,
        })
    }

    fn next_csi(&mut self, start_sequence: usize) -> Option<EscapeSequenceOffsets> {
        let (csi_open_index, csi_open_char) = self.chars.next().expect("to not be finished");
        debug_assert_eq!(csi_open_char, '[');

        let start_parameters: usize = csi_open_index + csi_open_char.len_utf8();

        // Keep iterating while within the range of `0x30-0x3F`.
        let mut start_intermediates: usize = start_parameters;
        if let Some((_, end)) = self.chars_take_while(|c| matches!(c, '\x30'..='\x3F')) {
            start_intermediates = end;
        }

        // Keep iterating while within the range of `0x20-0x2F`.
        let mut start_final_byte: usize = start_intermediates;
        if let Some((_, end)) = self.chars_take_while(|c| matches!(c, '\x20'..='\x2F')) {
            start_final_byte = end;
        }

        // Take the last char.
        let end_of_sequence = match self.chars.next() {
            None => start_final_byte,
            Some((i, c)) => i + c.len_utf8(),
        };

        Some(EscapeSequenceOffsets::CSI {
            start_sequence,
            start_parameters,
            start_intermediates,
            start_final_byte,
            end: end_of_sequence,
        })
    }

    fn next_nf(&mut self, start_sequence: usize) -> Option<EscapeSequenceOffsets> {
        let (nf_open_index, nf_open_char) = self.chars.next().expect("to not be finished");
        debug_assert!(matches!(nf_open_char, '\x20'..='\x2F'));

        let start: usize = nf_open_index;
        let mut end: usize = start;

        // Keep iterating while within the range of `0x20-0x2F`.
        match self.chars_take_while(|c| matches!(c, '\x20'..='\x2F')) {
            Some((_, i)) => end = i,
            None => {
                return Some(EscapeSequenceOffsets::NF {
                    start_sequence,
                    start,
                    end,
                })
            }
        }

        // Get the final byte.
        match self.chars.next() {
            Some((i, c)) => end = i + c.len_utf8(),
            None => {}
        }

        Some(EscapeSequenceOffsets::NF {
            start_sequence,
            start,
            end,
        })
    }
}

impl<'a> Iterator for EscapeSequenceOffsetsIterator<'a> {
    type Item = EscapeSequenceOffsets;
    fn next(&mut self) -> Option<Self::Item> {
        match self.chars.peek() {
            Some((_, '\x1B')) => self.next_sequence(),
            Some((_, _)) => self.next_text(),
            None => None,
        }
    }
}

/// Strips problematic ANSI escape sequences from a string.
///
/// Ideally, this will be replaced with something that uses [[Attributes]] to create a table of char offsets
/// -> absolute styles and style deltas. Something like that would let us simplify the printer (and support
/// re-printing OSC hyperlink commands).
pub fn strip_problematic_sequences(text: &str) -> String {
    use EscapeSequenceOffsets::*;

    let mut buffer = String::with_capacity(text.len());
    for seq in EscapeSequenceOffsetsIterator::new(text) {
        match seq {
            Text { start, end } => buffer.push_str(&text[start..end]),
            Unknown { start, end } => buffer.push_str(&text[start..end]),

            NF {
                start_sequence: start,
                start: _,
                end,
            } => buffer.push_str(&text[start..end]),

            CSI {
                start_sequence: start,
                start_parameters: _,
                start_intermediates: _,
                start_final_byte: _,
                end,
            } => buffer.push_str(&text[start..end]),

            OSC {
                start_sequence: _,
                start_command: _,
                start_terminator: _,
                end: _,
            } => {
                // TODO(eth-p): Support re-printing hyperlinks.
                // In the meantime, strip these.
            }
        }
    }

    buffer
}

#[cfg(test)]
mod tests {
    use crate::vscreen::{
        strip_problematic_sequences, EscapeSequenceOffsets, EscapeSequenceOffsetsIterator,
    };

    #[test]
    fn test_escape_sequence_offsets_iterator_parses_text() {
        let mut iter = EscapeSequenceOffsetsIterator::new("text");
        assert_eq!(
            iter.next(),
            Some(EscapeSequenceOffsets::Text { start: 0, end: 4 })
        );
    }

    #[test]
    fn test_escape_sequence_offsets_iterator_parses_text_stops_at_esc() {
        let mut iter = EscapeSequenceOffsetsIterator::new("text\x1B[ming");
        assert_eq!(
            iter.next(),
            Some(EscapeSequenceOffsets::Text { start: 0, end: 4 })
        );
    }

    #[test]
    fn test_escape_sequence_offsets_iterator_parses_osc_with_bel() {
        let mut iter = EscapeSequenceOffsetsIterator::new("\x1B]abc\x07");
        assert_eq!(
            iter.next(),
            Some(EscapeSequenceOffsets::OSC {
                start_sequence: 0,
                start_command: 2,
                start_terminator: 5,
                end: 6,
            })
        );
    }

    #[test]
    fn test_escape_sequence_offsets_iterator_parses_osc_with_st() {
        let mut iter = EscapeSequenceOffsetsIterator::new("\x1B]abc\x1B\\");
        assert_eq!(
            iter.next(),
            Some(EscapeSequenceOffsets::OSC {
                start_sequence: 0,
                start_command: 2,
                start_terminator: 5,
                end: 7,
            })
        );
    }

    #[test]
    fn test_escape_sequence_offsets_iterator_parses_osc_thats_broken() {
        let mut iter = EscapeSequenceOffsetsIterator::new("\x1B]ab");
        assert_eq!(
            iter.next(),
            Some(EscapeSequenceOffsets::OSC {
                start_sequence: 0,
                start_command: 2,
                start_terminator: 4,
                end: 4,
            })
        );
    }

    #[test]
    fn test_escape_sequence_offsets_iterator_parses_csi() {
        let mut iter = EscapeSequenceOffsetsIterator::new("\x1B[m");
        assert_eq!(
            iter.next(),
            Some(EscapeSequenceOffsets::CSI {
                start_sequence: 0,
                start_parameters: 2,
                start_intermediates: 2,
                start_final_byte: 2,
                end: 3
            })
        );
    }

    #[test]
    fn test_escape_sequence_offsets_iterator_parses_csi_with_parameters() {
        let mut iter = EscapeSequenceOffsetsIterator::new("\x1B[1;34m");
        assert_eq!(
            iter.next(),
            Some(EscapeSequenceOffsets::CSI {
                start_sequence: 0,
                start_parameters: 2,
                start_intermediates: 6,
                start_final_byte: 6,
                end: 7
            })
        );
    }

    #[test]
    fn test_escape_sequence_offsets_iterator_parses_csi_with_intermediates() {
        let mut iter = EscapeSequenceOffsetsIterator::new("\x1B[$m");
        assert_eq!(
            iter.next(),
            Some(EscapeSequenceOffsets::CSI {
                start_sequence: 0,
                start_parameters: 2,
                start_intermediates: 2,
                start_final_byte: 3,
                end: 4
            })
        );
    }

    #[test]
    fn test_escape_sequence_offsets_iterator_parses_csi_with_parameters_and_intermediates() {
        let mut iter = EscapeSequenceOffsetsIterator::new("\x1B[1$m");
        assert_eq!(
            iter.next(),
            Some(EscapeSequenceOffsets::CSI {
                start_sequence: 0,
                start_parameters: 2,
                start_intermediates: 3,
                start_final_byte: 4,
                end: 5
            })
        );
    }

    #[test]
    fn test_escape_sequence_offsets_iterator_parses_csi_thats_broken() {
        let mut iter = EscapeSequenceOffsetsIterator::new("\x1B[");
        assert_eq!(
            iter.next(),
            Some(EscapeSequenceOffsets::CSI {
                start_sequence: 0,
                start_parameters: 2,
                start_intermediates: 2,
                start_final_byte: 2,
                end: 2
            })
        );

        let mut iter = EscapeSequenceOffsetsIterator::new("\x1B[1");
        assert_eq!(
            iter.next(),
            Some(EscapeSequenceOffsets::CSI {
                start_sequence: 0,
                start_parameters: 2,
                start_intermediates: 3,
                start_final_byte: 3,
                end: 3
            })
        );

        let mut iter = EscapeSequenceOffsetsIterator::new("\x1B[1$");
        assert_eq!(
            iter.next(),
            Some(EscapeSequenceOffsets::CSI {
                start_sequence: 0,
                start_parameters: 2,
                start_intermediates: 3,
                start_final_byte: 4,
                end: 4
            })
        );
    }

    #[test]
    fn test_escape_sequence_offsets_iterator_parses_nf() {
        let mut iter = EscapeSequenceOffsetsIterator::new("\x1B($0");
        assert_eq!(
            iter.next(),
            Some(EscapeSequenceOffsets::NF {
                start_sequence: 0,
                start: 1,
                end: 4
            })
        );
    }

    #[test]
    fn test_escape_sequence_offsets_iterator_parses_nf_thats_broken() {
        let mut iter = EscapeSequenceOffsetsIterator::new("\x1B(");
        assert_eq!(
            iter.next(),
            Some(EscapeSequenceOffsets::NF {
                start_sequence: 0,
                start: 1,
                end: 1
            })
        );
    }

    #[test]
    fn test_escape_sequence_offsets_iterator_iterates() {
        let mut iter = EscapeSequenceOffsetsIterator::new("text\x1B[33m\x1B]OSC\x07\x1B(0");
        assert_eq!(
            iter.next(),
            Some(EscapeSequenceOffsets::Text { start: 0, end: 4 })
        );
        assert_eq!(
            iter.next(),
            Some(EscapeSequenceOffsets::CSI {
                start_sequence: 4,
                start_parameters: 6,
                start_intermediates: 8,
                start_final_byte: 8,
                end: 9
            })
        );
        assert_eq!(
            iter.next(),
            Some(EscapeSequenceOffsets::OSC {
                start_sequence: 9,
                start_command: 11,
                start_terminator: 14,
                end: 15
            })
        );
        assert_eq!(
            iter.next(),
            Some(EscapeSequenceOffsets::NF {
                start_sequence: 15,
                start: 16,
                end: 18
            })
        );
        assert_eq!(iter.next(), None);
    }

    #[test]
    fn test_strip_problematic_sequences() {
        assert_eq!(
            strip_problematic_sequences("text\x1B[33m\x1B]OSC\x1B\\\x1B(0"),
            "text\x1B[33m\x1B(0"
        );
    }
}
